// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Implementation of IEnumVariant for CCollection

#include "precomp.h"
#include "enumvar.h"

/*
 * CEnumVariant::Create
 *
 * Purpose:
 *  Creates an instance of the IEnumVARIANT enumerator object and initializes it.
 *
 * Parameters:
 *  psa        Safe array containing items to be enumerated.
 *  cElements  Number of items to be enumerated. 
 *  ppenumvariant    Returns enumerator object.
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT CEnumVariant::Create(CCollection FAR *pCollection
		, CEnumVariant FAR* FAR* ppenumvariant) 
{   
    CEnumVariant FAR* penumvariant = NULL;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppenumvariant, 0))){
    		AddInDebugOutput(NULL, TEXT("CEnumVariant::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
	    *ppenumvariant = NULL;
    
    	penumvariant = new CEnumVariant();
    	if (penumvariant == NULL){
	   		AddInDebugOutput(NULL
   				, TEXT("CEnumVariant::Create")
   				, IDS_DEBUG_IENUMVARIANTALLOCFAILED
   				, NULL); 
			return ResultFromScode(E_FAIL);
		}        

		penumvariant->m_pCollection = pCollection;

		penumvariant->m_pCurrentNode = pCollection->m_pFirstNode;
    
	    *ppenumvariant = penumvariant;
    
    	return S_OK;
	}

	#ifdef WIN32
	EXCEPT{
    	if (penumvariant == NULL)
        	return ResultFromScode(E_FAIL);   
                              
    	delete penumvariant;

    	return ResultFromScode(E_FAIL);
	}           
	#endif
}

/*
 * CEnumVariant::CEnumVariant
 *
 * Purpose:
 *  Constructor for CEnumVariant object. Initializes members to NULL.
 *
 */
CEnumVariant::CEnumVariant()
{    
	m_pCurrentNode = NULL;
	m_pCollection = NULL;
	m_cRef = 1;
}

/*
 * CEnumVariant::~CEnumVariant
 *
 * Purpose:
 *  Destructor for CEnumVariant object. 
 *
 */
CEnumVariant::~CEnumVariant()
{                   
}

/*
 * CEnumVariant::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
STDMETHODIMP
CEnumVariant::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
        
    if (iid == IID_IUnknown || iid == IID_IEnumVARIANT) 
        *ppv = this;     
    else return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CEnumVariant::AddRef(void)
{
    return ++m_cRef;  // AddRef Application Object if enumerator will outlive application object
}


STDMETHODIMP_(ULONG)
CEnumVariant::Release(void)
{
    if(--m_cRef == 0)
    {
        delete this;
        return 0;
    }
    return m_cRef;
}

/*
 * CEnumVariant::Next
 *
 * Purpose:
 *  Retrieves the next cElements elements. Implements IEnumVARIANT::Next. 
 *
 */
STDMETHODIMP
CEnumVariant::Next(ULONG cElements, VARIANT FAR* pvar, ULONG FAR* pcElementFetched)
{ 
    ULONG i=0;

	TRY{                  
		if(m_pCurrentNode == NULL)
			return ResultFromScode(S_FALSE);
	
		if(pcElementFetched)
    		*pcElementFetched = 0;
    
    	for (; i<cElements; i++){
        	VariantInit(&pvar[i]);
			pvar[i].vt = VT_DISPATCH;			
		}

	    // Retrieve the next cElements elements.
		for(i=0; m_pCurrentNode && i < cElements 
				&& i < m_pCollection->m_cElements
				; m_pCurrentNode = m_pCurrentNode->m_pNext, i++){
		
			pvar[i].vt = VT_DISPATCH;
			pvar[i].pdispVal = m_pCurrentNode->m_pDispatch;
			(pvar[i].pdispVal)->AddRef();
			
			if(pcElementFetched){
				(*pcElementFetched)++;
			}
		}

		return S_OK;
	}

	#ifdef _WIN32    
	EXCEPT{
    	for (i=0; i<cElements; i++){
        	VariantClear(&pvar[i]);
		}

    	return ResultFromScode(E_FAIL);    
	}
	#endif
}

/*
 * CEnumVariant::Skip
 *
 * Purpose:
 *  Skips the next cElements elements. Implements IEnumVARIANT::Skip. 
 *
 */
STDMETHODIMP
CEnumVariant::Skip(ULONG cElements)
{   
	ULONG i = 0;
	CNode FAR *pNode;

	TRY{    
	    // Retrieve the next cElements elements.
		for(;pNode && i < cElements && i < m_pCollection->m_cElements
				; pNode = pNode->m_pNext, i++) {}

		if(i < cElements) {
			return ResultFromScode(E_FAIL);
		}else{
			m_pCurrentNode = pNode;

			return S_OK;
		}				
	}

	#ifdef _WIN32    
	EXCEPT{
    	return ResultFromScode(E_FAIL);    
	}
	#endif
}

/*
 * CEnumVariant::Reset
 *
 * Purpose:
 *  Resets the current element in the enumerator to the beginning. Implements IEnumVARIANT::Reset. 
 *
 */
STDMETHODIMP
CEnumVariant::Reset()
{ 
	m_pCurrentNode = m_pCollection->m_pFirstNode;
    
    return ResultFromScode(E_FAIL);
}

/*
 * CEnumVariant::Clone
 *
 * Purpose:
 *  Creates a copy of the current enumeration state. Implements IEnumVARIANT::Clone. 
 *
 */
STDMETHODIMP
CEnumVariant::Clone(IEnumVARIANT FAR* FAR* ppenum)
{
    return ResultFromScode(E_FAIL);        
}


