// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Hash table for collections tyo look up by ID or Name very
//		quickly.

#include "precomp.h"
#include <stdlib.h>


/*
 * CHashTable::Create
 *
 * Purpose:
 *  Creates an instance of the client object and initializes it. Use of
 *	the class should never call new.
 *
 * Parameters:
 *  bSymbol    		Are we a Symbol Hash table or a ID Hash Table
 *	ppHashTable	Pointer to return created symbol table
 *	
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CHashTable::Create(BOOL bSymbol, CHashTable FAR* FAR* ppHashTable){
	CHashTable FAR* pHashTable;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppHashTable, 0))){
    		AddInDebugOutput(NULL, TEXT("CHashTable::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
	    *ppHashTable = NULL;
		
		pHashTable = new CHashTable(bSymbol);
	
		if(pHashTable == NULL){
			AddInDebugOutput(NULL
					, TEXT("CHashTable::Create")
					, IDS_DEBUG_CREATEHASHTABLEFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}
	    
		pHashTable->m_pHash = (CNode FAR * FAR*) GlobalAllocPtr(GHND
			, TABLE_SIZE * sizeof(CNode FAR *));

		if(FAILED(CheckPointer(pHashTable->m_pHash))){
			AddInDebugOutput(NULL
					, TEXT("CHashTable::Create")
					, IDS_DEBUG_CREATEHASHTABLEFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}

		*ppHashTable = pHashTable;               
                
		return S_OK;	
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CHashTable::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);		
	}
}    
    
/*
 * CHashTable::CHashTable
 *
 * Purpose:
 *  Constructor
 *
 */ 

CHashTable::CHashTable(BOOL bSymbol){
	m_pHash = NULL;
	m_bSymbol = bSymbol;
	m_dwSymbolCount = 0;
	m_dwIDCount = 0;
}	
	
/*
 * CHashTable::~CHashTable
 *
 * Purpose:
 *  Destructor
 *
 */ 

CHashTable::~CHashTable(){
	if(SUCCEEDED(CheckPointer(m_pHash)))
		GlobalFreePtr(m_pHash);
}	
	
/*
 * CHashTable::MakeHash
 *
 * Purpose:
 *  Transforms a string into a hash tables's index.
 *
 * Parameters:
 *  pszKey    		String to convert
 *
 * Return Value:
 *  int
 *
 */ 

int CHashTable::MakeHash(LPTSTR pszKey){
	int sum = 0;
	int ret = 0;

	TRY{		
		// Add em up
		while(*pszKey){
			sum += (UINT) *pszKey;
			pszKey++;	
		}
	
		return sum % TABLE_SIZE;
	}

	#ifdef _WIN32
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CHashTable::MakeHash")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return 0;
	}
	#endif
			
}	

/*
 * CHashTable::FindSymbol
 *
 * Purpose:
 *  Find the Symbol in hash table
 *
 * Parameters:
 *  pszKey    		String to convert
 *
 * Return Value:
 *  int
 *
 */ 

HRESULT CHashTable::Find(LPTSTR pszKey, CNode FAR* FAR* ppNode){
	int posn = 0;
	CNode FAR *pNode = NULL;
	TCHAR buf[MAX_DIGITS]; // for converting numbers to strings
	           
	TRY{	
		*ppNode = NULL;
		
		posn = MakeHash(pszKey);
		
		if(posn >= TABLE_SIZE){
			return ResultFromScode(E_FAIL);
		} 
	                        
	    // If symbol hash table
		if(m_bSymbol){
			if(m_dwSymbolCount == 0){
				return ResultFromScode(E_FAIL);
			}
		
			// If clash occurs link tem
			for(pNode = m_pHash[posn]; pNode != NULL; pNode = pNode->m_pNextSymbol){
				if(!lstrcmpi(pszKey, (LPTSTR) pNode->m_bstrName)){
					*ppNode = pNode;
					return S_OK;
				}
			}

			return ResultFromScode(E_FAIL);

	    // If ID hash table
		}else{
			if(m_dwIDCount == 0){
				return ResultFromScode(E_FAIL);
			}
			
			// If clash occurs link tem
			for(pNode = m_pHash[posn]; pNode != NULL; pNode = pNode->m_pNextID){
				#ifdef _UNICODE
					_ultow(pNode->m_dwID, buf, 10);
				#else
					_ultoa(pNode->m_dwID, buf, 10);
				#endif
			
				if(!lstrcmpi(pszKey, (LPTSTR) buf)){
					*ppNode = pNode;
					return S_OK;
				}
			}

			return ResultFromScode(E_FAIL);
		}
	}

	#ifdef _WIN32
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CHashTable::Find")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL); 	
	}
	#endif
}


/*
 * CHashTable::InsertSymbol
 *
 * Purpose:
 *  Find the Symbol in hash table
 *
 * Parameters:
 *  pszKey    		String to insert
 *
 * Return Value:
 *  int
 *
 */ 

HRESULT CHashTable::Insert(LPTSTR pszKey, CNode FAR *pNode){
	CNode FAR *pLocalNode;
	int posn;

	TRY{	
		posn = MakeHash(pszKey);
		
		if(posn >= TABLE_SIZE){
			return ResultFromScode(E_FAIL);
		} 
		
		// If is symbol hash table	
		if(m_bSymbol){
			if(m_pHash[posn] == NULL){
				m_pHash[posn] = pNode;
			}else{
				pLocalNode = m_pHash[posn];
				m_pHash[posn] = pNode;
				pNode->m_pNextSymbol = pLocalNode;
			}
			
			m_dwSymbolCount++;
			
		// If is symbol hash table	
		}else{
			if(m_pHash[posn] == NULL){
				m_pHash[posn] = pNode;
			}else{
				pLocalNode = m_pHash[posn];
				m_pHash[posn] = pNode;
				pNode->m_pNextID = pLocalNode;
			}
			
			m_dwIDCount++;
		}

		return S_OK;
	}

	#ifdef _WIN32
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CHashTable::Insert")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);		
	}
	#endif
}

/*
 * CHashTable::Remove
 *
 * Purpose:
 *  Remove the Symbol from hash table
 *
 * Parameters:
 *  pszKey    		String to remove
 *
 * Return Value:
 *  int
 *
 */ 

HRESULT CHashTable::Remove(LPTSTR pszKey){
	int posn = 0;
	CNode FAR *pPrevNode;
	CNode FAR *pNode = NULL;
	TCHAR buf[MAX_DIGITS]; // for converting numbers to strings

	TRY{
		if(m_bSymbol){
			posn = MakeHash(pszKey);
		
			for(;pNode!=NULL;pPrevNode=pNode,pNode=pNode->m_pNextSymbol){
				if(!lstrcmp(pszKey, (LPTSTR) pNode->m_bstrName)){
					break;
				}
			}             
	
			if(pNode == NULL){
				return ResultFromScode(E_FAIL);		
			}			
	
			if(pPrevNode == NULL){ //First Row
				m_pHash[posn] = pNode->m_pNextSymbol;
			}else{
				pPrevNode->m_pNextSymbol = pNode->m_pNextSymbol;
			}
			
			m_dwSymbolCount--;
		}else{
			posn = MakeHash(pszKey);
		
			for(;pNode!=NULL;pPrevNode=pNode,pNode=pNode->m_pNextID){
				#ifdef _UNICODE
					_ultow(pNode->m_dwID, buf, 10);
				#else
					_ultoa(pNode->m_dwID, buf, 10);
				#endif
			
				if(!lstrcmpi(pszKey, (LPTSTR) buf)){
					return S_OK;
				}
			}             
	
			if(pNode == NULL){
				return ResultFromScode(E_FAIL);		
			}			
	
			if(pPrevNode == NULL){ //First Row
				m_pHash[posn] = pNode->m_pNextID;
			}else{
				pPrevNode->m_pNextID = pNode->m_pNextID;
			}

			m_dwIDCount--;
		}
	
		return S_OK;
	}

	#ifdef _WIN32
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CHashTable::Remove")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
	#endif
} 




